﻿using System;
using System.Drawing;
using System.Drawing.Imaging;

namespace DemoApp.Filters
{
    /// <summary>
    /// Floyd-Steinberg dither filter (color)
    /// </summary>
    public class FloydSteinbergDitherColorFilter
    {
        /// <summary>
        /// Number of levels
        /// </summary>
        private int levels;

        /// <summary>
        /// Number of levels [2:255]
        /// </summary>
        public int Levels
        {
            get { return levels; }
            set
            {
                if (value < 2) levels = 2;
                else if (value > 255) levels = 255;
                else levels = value;
            }
        }

        /// <summary>
        /// Constructor with levels parameter
        /// </summary>
        /// <param name="levels">Number of levels [2:255]</param>
        public FloydSteinbergDitherColorFilter(int levels)
        {
            Levels = levels;
        }

        /// <summary>
        /// Apply filter, the original image will be modified.
        /// </summary>
        /// <param name="image">The filter is applied on this image</param>
        public void ApplyInPlace(Bitmap image)
        {
            // Lock bits
            BitmapData bmd = image.LockBits(new Rectangle(0, 0, image.Width, image.Height),
                ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);
            int nVal;
            int w_3 = image.Width * 3; // Width of a row
            int h = image.Height; // Image height
            int i, j;
            double intervalSize = 255.0 / (levels - 1); // Size of an interval
            int roundedColor; // Color rounded to the nearest color level
            float quant_err;
            float[,] q_err = new float[w_3, h];
            for (i = 0; i < w_3; i++) for (j = 0; j < h; j++) q_err[i, j] = 0;
            unsafe
            {
                // Iterate through rows
                for (j = 0; j < h; ++j)
                {
                    // Get row
                    byte* row = (byte*)bmd.Scan0 + (j * bmd.Stride);
                    // Iterate through columns
                    for (i = 0; i < w_3; ++i)
                    {
                        // Add quantization error
                        nVal = (int)(row[i] + q_err[i, j]);
                        if (nVal > 255) nVal = 255;
                        else if (nVal < 0) nVal = 0;
                        // Get rounded color
                        roundedColor = (int)(Math.Round(nVal / intervalSize) * intervalSize);
                        // Calculate quantization error
                        quant_err = nVal - roundedColor;
                        // Sum quantization error using the following matrix:
                        // 1/16 * |  p 7|
                        //        |3 5 1|
                        // (p is the actual pixel)
                        if (i < w_3 - 3) q_err[i + 3, j] += quant_err * 0.4375f;
                        if (j < h - 1)
                        {
                            if (i > 2) q_err[i - 3, j + 1] += quant_err * 0.1875f;
                            q_err[i, j + 1] += quant_err * 0.3125f;
                            if (i < w_3 - 3) q_err[i + 3, j + 1] += quant_err * 0.0625f;
                        }
                        // Replace color
                        row[i] = (byte)roundedColor;
                    }
                }
            }
            image.UnlockBits(bmd); // Unlock bits
        }
    }
}
