﻿using System.Drawing;
using System.Drawing.Imaging;

namespace DemoApp.Filters
{
    /// <summary>
    /// Oil paint filter
    /// </summary>
    public class OilPaintFilter
    {
        /// <summary>
        /// Brush radius
        /// </summary>
        private int radius;

        /// <summary>
        /// Brush radius [1;...]
        /// </summary>
        public int Radius
        {
            get { return radius; }
            set
            {
                if (value < 1) radius = 1;
                else radius = value;
            }
        }

        /// <summary>
        /// Constructor with brush radius
        /// </summary>
        /// <param name="radius">Brush radius</param>
        public OilPaintFilter(int radius)
        {
            this.Radius = radius;
        }

        /// <summary>
        /// Apply filter, the original image will be modified.
        /// </summary>
        /// <param name="image">The filter is applied on this image</param>
        public void ApplyInPlace(Bitmap image)
        {
            // Clone image (the clone won't be modified)
            Bitmap original = (Bitmap)image.Clone();
            // Lock bits
            BitmapData bmd = image.LockBits(new Rectangle(0, 0, image.Width, image.Height),
                ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);
            BitmapData bmdOrg = original.LockBits(new Rectangle(0, 0, original.Width, original.Height),
                ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);

            int img_w = image.Width, img_h = image.Height;
            int stride = bmd.Stride;
            int w_3 = image.Width * 3; // Width of a row
            int radius_3 = radius * 3;
            int i, j, k, l, avg, idx, max;
            int[] red = new int[256];
            int[] green = new int[256];
            int[] blue = new int[256];
            int[] intensities = new int[256];

            unsafe
            {
                // Iterate through rows
                for (j = 0; j < img_h; ++j)
                {
                    // Get rows
                    byte* row = (byte*)bmd.Scan0 + (j * stride);
                    byte* rowOrg = (byte*)bmdOrg.Scan0 + (j * stride);

                    // Clear arrays
                    for (k = 0; k < 256; ++k) red[k] = green[k] = blue[k] = intensities[k] = 0;

                    // Build up arrays with the surrounding of the first column of the actual row
                    // Horizontal
                    for (k = 0; k <= radius && k < img_w; ++k)
                    {
                        // Vertically from -radius to +radius
                        for (l = j < radius ? -j : -radius; j + l < img_h && l <= radius; ++l)
                        {
                            // Calculate index (relative to current row)
                            idx = l * stride + k * 3;
                            // Get luminance
                            avg = (int)(0.299 * rowOrg[idx + 2] + 0.587 * rowOrg[idx + 1] + 0.114 * rowOrg[idx]);
                            // Increase values
                            intensities[avg]++;
                            // Sum values
                            red[avg] += rowOrg[idx + 2];
                            green[avg] += rowOrg[idx + 1];
                            blue[avg] += rowOrg[idx];
                        }
                    }
                    // Get the most frequent intensity
                    max = 0;
                    for (k = 0; k < 256; k++) if (intensities[max] < intensities[k]) max = k;
                    // First element of the row will be the average of the maximal intensity
                    row[2] = (byte)(red[max] / intensities[max]);
                    row[1] = (byte)(green[max] / intensities[max]);
                    row[0] = (byte)(blue[max] / intensities[max]);

                    // Iterate through other columns
                    // Rather than building up the whole array again, we just take in another column
                    // (+radius) and remove a column (-radius-1)
                    for (i = 3; i < w_3; i += 3)
                    {
                        // If we can remove a column
                        if (i / 3 - radius - 1 >= 0)
                        {
                            // Remove column (vertically from -radius to +radius)
                            for (l = j < radius ? -j : -radius; j + l < img_h && l <= radius; ++l)
                            {
                                // Calculate index (relative to current row)
                                idx = l * stride + i - radius_3 - 3;
                                // Get luminance
                                avg = (int)(0.299 * rowOrg[idx + 2] + 0.587 * rowOrg[idx + 1] + 0.114 * rowOrg[idx]);
                                // Decrease values
                                intensities[avg]--;
                                // Sum values
                                red[avg] -= rowOrg[idx + 2];
                                green[avg] -= rowOrg[idx + 1];
                                blue[avg] -= rowOrg[idx];
                            }
                        }
                        // If we can add a new column
                        if (i / 3 + radius < img_w)
                        {
                            // Add column (vertically from -radius to +radius)
                            for (l = j < radius ? -j : -radius; j + l < img_h && l <= radius; ++l)
                            {
                                // Calculate index (relative to current row)
                                idx = l * stride + i + radius_3;
                                // Get luminance
                                avg = (int)(0.299 * rowOrg[idx + 2] + 0.587 * rowOrg[idx + 1] + 0.114 * rowOrg[idx]);
                                // Increase values
                                intensities[avg]++;
                                // Sum values
                                red[avg] += rowOrg[idx + 2];
                                green[avg] += rowOrg[idx + 1];
                                blue[avg] += rowOrg[idx];
                            }
                        }
                        // Get the most frequent intensity
                        max = 0;
                        for (k = 0; k < 256; k++) if (intensities[max] < intensities[k]) max = k;
                        // Actual element will be the average of the maximal intensity
                        row[i + 2] = (byte)(red[max] / intensities[max]);
                        row[i + 1] = (byte)(green[max] / intensities[max]);
                        row[i] = (byte)(blue[max] / intensities[max]);
                    }
                }
            }
            image.UnlockBits(bmd); // Unlock bits
            original.UnlockBits(bmdOrg); // Unlock bits
            original.Dispose(); // Dispose the copied image
        }
    }
}
